from typing import Union, List
import numpy as np
import random,sys,os
import torch

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from log import logging
logger=logging.getLogger(__name__)
from .base_reader import convert_to_tensor
SPECIAL_TOKEN_NUMS=2#[CLS] [SEP]

class InputExample:
    """
    Structure for one input example with texts, the label and a unique id
    """
    def __init__(self, guid: str = '', text_list: List[str] = None,  label = 0):
        """
        text_list是一个有两个str的list
        """
        self.guid = guid
        self.text_list = text_list
        self.label = label

    def __str__(self):
        return "<InputExample> label: {}, text pairs : {}".format(str(self.label), "; ".join(self.text_list))

class InputFeatures:
    def __init__(self, input_ids,token_type_ids,attention_mask):
        self.input_ids=input_ids
        self.token_type_ids=token_type_ids
        self.attention_mask=attention_mask
    def __str__(self) -> str:
        return "<InputFeature> input_ids: {}\n token_type_ids: {}\n attention_mask: {}\n".format(' '.join([str(x) for x in self.input_ids]),
                                                                                            ' '.join([str(x) for x in self.token_type_ids]),
                                                                                            ' '.join([str(x) for x in self.attention_mask]))



def getExamples(file_path,label2id,isCL=False,mode="train",sentence1_idx=0,sentence2_idx=1,label_idx=2,filter_heads=False):
    with open(file_path, encoding='utf-8') as f:
        lines = f.readlines()
    if filter_heads:
        logger.info("Heads like : {}".format(lines[0]))
        lines=lines[1:]
    train_data=[]
    for line in lines:
        line_split = line.strip().split('\t')
        sentence1 = line_split[sentence1_idx]
        sentence2 = line_split[sentence2_idx]
        label = line_split[label_idx]
        if isCL==True:
            if str(label)=='1':
                train_data.append(InputExample(text_list=[sentence1,sentence2],label=label2id[label]))
        else:
            train_data.append(InputExample(text_list=[sentence1,sentence2],label=label2id[label]))
    logger.info("*****************************Logging some {} examples*****************************".format(mode))
    logger.info("Total {} nums is : {}".format(mode,len(train_data)))
    for _ in range(5):
        i=random.randint(0,len(train_data)-1)
        logger.info("\t".join(train_data[i].text_list)+"\t"+str(train_data[i].label))
    return train_data


def convert_examples_to_features(examples, tokenizer, max_seq_len=None):
    '''
    return like [{'input_ids':tensor,'attention_mask':tensor,'token_type_ids':tensor},{'input_ids':tensor,'attention_mask':tensor,'token_type_ids':tensor}]
    max_seq_len means the max seq_len for single sentence, not pair sentence
    '''
    def get_Feature_from_sentence(sentence,max_len):
        tokens = []
        token_type_ids = []
        tokens.append("[CLS]")
        token_type_ids.append(0)
        
        tokens_text_a = tokenizer.tokenize(sentence)
        tokens += tokens_text_a
        token_type_ids += [0] * len(tokens_text_a)
        tokens.append('[SEP]')
        token_type_ids.append(0)
        if len(tokens) >= max_len:
            tokens = tokens[:max_len]
            token_type_ids = token_type_ids[:max_len]

        seq_len = len(tokens)
        pad_len = max_len - seq_len
        # print(tokens)
        input_ids = tokenizer.convert_tokens_to_ids(tokens)
        input_ids += [0] * pad_len
        attention_mask = [1] * seq_len + [0] * pad_len
        token_type_ids += [0] * pad_len
        assert len(input_ids) == len(token_type_ids) == len(attention_mask) == max_len
        features=InputFeatures(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask)
        return features

    if isinstance(examples[0],list):
        assert type(examples[0][0])==type(examples[0][1])==str
        examples=[InputExample(text_list=text_list) for text_list in examples]
        #传进来的每一个值是list不是Example类型
        
    features_of_a = []
    features_of_b = []
    labels=[]
    assert len(examples[0].text_list)==2
    max_len_of_1=0
    max_len_of_2=0

    for example in examples:
        length_of_1=len(tokenizer.tokenize(example.text_list[0]))
        length_of_2=len(tokenizer.tokenize(example.text_list[1]))
        if length_of_1+SPECIAL_TOKEN_NUMS>max_len_of_1:
            max_len_of_1=length_of_1
        if length_of_2+SPECIAL_TOKEN_NUMS>max_len_of_2:
            max_len_of_2=length_of_2     

    if max_seq_len!=None:
        max_len_of_1=min(max_seq_len,max_len_of_1)
        max_len_of_2=min(max_seq_len,max_len_of_2)

    for example_index, example in enumerate(examples):
        feature_of_a=get_Feature_from_sentence(sentence=example.text_list[0],max_len=max_len_of_1)
        feature_of_b=get_Feature_from_sentence(sentence=example.text_list[1],max_len=max_len_of_2)
        
        features_of_a.append(feature_of_a)
        features_of_b.append(feature_of_b)
        labels.append(example.label)
    
    features_of_a=convert_to_tensor(features=features_of_a)
    features_of_b=convert_to_tensor(features=features_of_b)
    #{"input_ids":batch_size,max_seq_len,...}
    labels=torch.LongTensor(labels)
    return features_of_a,features_of_b,labels


def convert_sentences_to_features(sentences,tokenizer,max_seq_len=None):
    assert type(sentences)==list
    tokenized_sentences=[tokenizer.tokenize(sen) for sen in sentences]
    max_len=max([len(sen)+2 for sen in tokenized_sentences])
    if max_seq_len!=None:
        max_len=min(max_seq_len,max_len)

    all_sen_features=[]
    for tokenized_sen in tokenized_sentences:
        tokens = []
        token_type_ids = []
        tokens.append("[CLS]")
        token_type_ids.append(0)

        tokens += tokenized_sen
        token_type_ids += [0] * len(tokenized_sen)
        tokens.append('[SEP]')
        token_type_ids.append(0)
        if len(tokens) >= max_len:
            tokens = tokens[:max_len]
            token_type_ids = token_type_ids[:max_len]

        seq_len = len(tokens)
        pad_len = max_len - seq_len
        # print(tokens)
        input_ids = tokenizer.convert_tokens_to_ids(tokens)
        input_ids += [0] * pad_len
        attention_mask = [1] * seq_len + [0] * pad_len
        token_type_ids += [0] * pad_len
        assert len(input_ids) == len(token_type_ids) == len(attention_mask) == max_len
        features=InputFeatures(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask)
        all_sen_features.append(features)
    
    return convert_to_tensor(all_sen_features)#{"input_ids":tensor,...}


import logging
import os
import csv
from typing import List
from sklearn import metrics
import numpy as np
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from log import logging
#logging.root.setLevel(logging.INFO)

logger = logging.getLogger(__name__)

class clsEvaluator:
    """
    This evaluator can be used with the CrossEncoder class.
    It is designed for CrossEncoders with 2 or more outputs. It measure the
    accuracy of the predict class vs. the gold labels.
    """
    def __init__(self, sentences, label_ids, label2id, name: str='', write_csv: bool = True):
        self.sentences = sentences
        if type(label_ids[0])==str:
            logger.info("input label is str type, converting to id!")
            label_ids=[label2id[label_id] for label_id in label_ids]
        self.label_ids = label_ids
        self.label2id=label2id
        self.id2label={id_:tag for tag,id_ in self.label2id.items()}
        logger.info('label2id like : {}'.format(self.label2id))
        self.name = name
        self.label2metrics={}
        for key in self.label2id.keys():
            self.label2metrics[key]={'golden_num':0,'predict_num':0,'predict_correct_num':0}
            #标签是key的真实样本的数量、预测key的数量、预测正确的样本数量
        for tag_id in self.label_ids:
            tag=self.id2label[tag_id]
            self.label2metrics[tag]['golden_num']+=1

        for tag in self.label2metrics.keys():
            logger.info('The number of '+tag+' in dataset is {}'.format(self.label2metrics[tag]['golden_num']))

        self.csv_file = "ClsEvaluator" + ("_" + name if name else '') + "_results.csv"
        self.csv_headers = ["epoch", "steps", "Accuracy"]
        self.write_csv = write_csv

        logging.info("Evalautor sentence like : \n")
        for i in range(5):
            logging.info(self.sentences[i]+"\t"+str(self.label_ids[i])+"\n")

    def __call__(self, model, label2id={'0':0,'1':1}, batch_size=32, output_path: str = None, epoch: int = -1, steps: int = -1) -> float:
        '''
        model.predict根据传进去的列表得到dataloader，然后dataloader.collate_fn==smart_collate_fn
        而在smart_collate_fn中做的是convert_examples_to_examples，
        convert_examples_to_ids中，如果examples中的元素不是InputExample类型，那么会转换为Example类型
        '''
        if epoch != -1:
            if steps == -1:
                out_txt = " after epoch {}:".format(epoch)
            else:
                out_txt = " in epoch {} after {} steps:".format(epoch, steps)
        else:
            out_txt = ":"

        logging.info("ClsEvaluator: Evaluating the model on " + self.name + " dataset" + out_txt)

        pred_scores = model.predict(is_pairs=False, dataloader=self.sentences, batch_size=batch_size)#(num_eval_examples,num_classes)
        pred_label_ids = np.argmax(pred_scores, axis=1)
        assert len(pred_label_ids) == len(self.label_ids)
        acc = np.sum(pred_label_ids == self.label_ids) / len(self.label_ids)

        logging.info("Accuracy: {:.3f}".format(acc))

        for i,pred_tag_id in enumerate(pred_label_ids):
            tag=self.id2label[pred_tag_id]
            self.label2metrics[tag]['predict_num']+=1
            if pred_tag_id==self.label_ids[i]:
                self.label2metrics[tag]['predict_correct_num']+=1

        for label,values in self.label2metrics.items():
            precision=values['predict_correct_num']/values['predict_num']
            recall=values['predict_correct_num']/values['golden_num']
            f1_score=2*precision*recall/(precision+recall)
            logger.info(label+'\t'+'precision : {}'+' recall : {}'+" f1 score : {}".format(precision,recall,f1_score))

        if len(label2id)==2:
            assert label2id=={"0":0,"1":1}
            y=np.array([label2id[tag] if type(tag)==str else tag for tag in self.label_ids])
            pred=pred_scores[:,1]
            fpr, tpr, thresholds = metrics.roc_curve(y, pred, pos_label=1)
            auc_val=metrics.auc(fpr, tpr)
            logging.info("AUC: {:.3f}".format(auc_val))

        if output_path is not None and self.write_csv:
            csv_path = os.path.join(output_path, self.csv_file)
            output_file_exists = os.path.isfile(csv_path)
            if len(label2id)==2 and "Auc" not in self.csv_headers:
                self.csv_headers.append("Auc")
            with open(csv_path, mode="a" if output_file_exists else 'w', encoding="utf-8") as f:
                writer = csv.writer(f)
                if not output_file_exists:
                    writer.writerow(self.csv_headers)

                if len(label2id)==2:
                    writer.writerow([epoch, steps, acc, auc_val])
                else:
                    writer.writerow([epoch, steps, acc])#AUC不能作为多分类的指标

        if len(label2id)==2:
            return auc_val
        else:
            return acc

    @classmethod
    def getAcc(self,predictions,labels):
        return sum(predictions==labels)/len(predictions)


import json,logging
from os import path
import torch
from typing import List,Optional

from .base_reader import convert_to_tensor

class InputExample:
    def __init__(self,guid='', seq_in: list=None, seq_out: list=None):
        '''
        为了避免混淆tokenize后用' '.join得到的str和普通的没有空格的str，强制传入了seq_in是list
        '''
        self.guid=guid
        assert type(seq_in)==list
        if seq_out is None:
            seq_out=['O']*len(seq_in)
        assert type(seq_out)==list
        assert len(seq_in)==len(seq_out)
        self.seq_in=seq_in
        self.seq_out=seq_out
    def __str__(self) -> str:
        return "<InputExample> seq_in: {}\n seq_out: {}\n".format(' '.join(self.seq_in),' '.join(self.seq_out))

class InputFeatures:
    def __init__(self, 
                input_ids=[101,2587,2587,2587,102],
                token_type_ids=[0,0,0,0,0],
                attention_mask=[1,1,1,1,1],
                label_ids=[0,0,0,0,0]):
        self.input_ids=input_ids
        self.token_type_ids=token_type_ids
        self.attention_mask=attention_mask
        self.label_ids=label_ids
    def __str__(self) -> str:
        return "<InputFeature> input_ids: {}\n token_type_ids: {}\n attention_mask: {}\n label_ids: {}\n".format(' '.join([str(x) for x in self.input_ids]),
                                                                                                                ' '.join([str(x) for x in self.token_type_ids]),
                                                                                                                ' '.join([str(x) for x in self.attention_mask]),
                                                                                                                ' '.join([str(x) for x in self.label_ids]))


def generate_SeqIn_and_SeqOut_based_Examples(example,tokenizer):
    '''
    example形如
    {"text": "五一劳动节是农历几时", "intent": "Calendar-Query", "slots": {"datetime_date": "五一劳动节"}}
    {"text": "快帮我查一下明天是星期几", "intent": "Calendar-Query", "slots": {"datetime_date": "明天"}}
    {"text": "看一下2019年有哪些好看的动作片", "intent": "Film-Query", "slots": {"year": "2019年"}}
    '''
    text=example["text"]
    slots=example["slots"]
    seq_in=tokenizer.tokenize(text)
    seq_out=['O' for _ in range(len(seq_in))]
    assert len(seq_out)==len(seq_in)
    need_print=False
    for slot_name,slot_values in slots.items():
        if type(slot_values)==str:
            slot_values=[slot_values]
        multiple_slots_in_text=False
        multiple_slot_begin_pos=[]
        multiple_slot_end_pos=[]
        for slot_value in slot_values:
            slot_value_tokens=tokenizer.tokenize(slot_value)
            slot_value_length=len(slot_value_tokens)
            slot_begin_pos=-1
            slot_end_pos=-1
            if len(text)==len(seq_in):
                #没有数字和英文字母出现
                assert slot_value_length==len(slot_value)
                slot_begin_pos=text.find(slot_value)
                if slot_begin_pos!=-1:
                    slot_end_pos=slot_begin_pos+slot_value_length
                    if text.count(slot_value)>1:
                        #也就是说除去slot_begin_pos位置后面还有当前的slot_value
                        new_text=text[slot_end_pos:]
                        new_pointer=slot_end_pos
                        while new_text.find(slot_value)!=-1:
                            #print(new_text,new_text.find(slot_value)+slot_value_length+new_pointer,new_pointer)
                            multiple_slot_begin_pos.append(new_text.find(slot_value)+new_pointer)
                            multiple_slot_end_pos.append(new_text.find(slot_value)+slot_value_length+new_pointer)
                            new_text=text[new_text.find(slot_value)+slot_value_length+new_pointer:]
                            new_pointer=text.find(new_text)
                            multiple_slots_in_text=True
            else:
                #有数字或者英文字母甚至是空格的出现导致tokenize后的长度不一致
                #text=看一下2019年有哪些好看的动作片, seq_in=看 一 下 2019 年 有 哪 些 好 看 的 动 作 片
                slot_begin_char=slot_value_tokens[0]
                for i,token in enumerate(seq_in):
                    if token==slot_begin_char:
                        char=''
                        if i+slot_value_length>len(seq_in):
                            print("槽值在句子中的开始位置是%d，槽值的长度是%d，而句子的长度是%d，槽位中槽值的结束位置超出了句子的长度"%(i,slot_value_length,len(seq_in)))
                            print("出现的问题的example是 ",example)
                            break
                        for j in range(i,i+slot_value_length):
                            if seq_in[j][:2]=='##':
                                char+=seq_in[j][2:]#出现英文单词
                            else:
                                char+=seq_in[j]
                        if tokenizer.do_lower_case:
                            slot_value=slot_value.lower()#大小写问题
                        if char==slot_value:
                            slot_begin_pos=i
                            slot_end_pos=i+slot_value_length
                            break
            
            if slot_begin_pos!=-1:
                seq_out[slot_begin_pos]='B-'+slot_name
                for pos in range(slot_begin_pos+1,slot_end_pos):
                    seq_out[pos]='I-'+slot_name
                    
        if multiple_slots_in_text:
            for start_pos,end_pos in zip(multiple_slot_begin_pos,multiple_slot_end_pos):
                seq_out[start_pos]='B-'+slot_name
                for pos in range(start_pos+1,end_pos):
                    seq_out[pos]='I-'+slot_name
            # print("text中出现了多个槽值，标注后的形式为:")
            # print(seq_in)
            # print(seq_out)
            # print(example)
            # print('-'*100)
            need_print=True
    if need_print:
        print(seq_in)
        print(seq_out)
        print(example)
        print('-'*100)
    return seq_in,seq_out

def write_seqIn_and_seqOut_from_original_file(seq_in_path,seq_out_path,file_path,tokenizer):
    examples=[]
    with open(file_path) as f:
        lines=f.readlines()
    for line in lines:
        examples.append(json.loads(line.strip()))
    
    f_seq_in=open(seq_in_path,"w",encoding='utf-8')
    f_seq_out=open(seq_out_path,"w",encoding="utf-8")

    for example in examples:
        seq_in,seq_out=generate_SeqIn_and_SeqOut_based_Examples(example,tokenizer)
        assert len(seq_in)==len(seq_out)
        f_seq_in.write(' '.join(seq_in)+"\n")
        f_seq_out.write(' '.join(seq_out)+'\n')
    f_seq_in.close()
    f_seq_out.close()


def getSeqIn_and_SeqOut(seq_in_path,seq_out_path,intent_path=None):
    def readData(path):
        data=[]
        with open(path,encoding="utf-8") as f:
            lines=f.readlines()
            for line in lines:
                data.append(line.strip().split(' '))
        return data

    seq_in=readData(path=seq_in_path)
    seq_out=readData(path=seq_out_path)
    if intent_path!=None:
        intents=readData(path=intent_path)
        return seq_in,seq_out,intents
    else:
        return seq_in,seq_out
    

def readExample(seq_in_list,seq_out_list=None):
    '''
    传进来的seq_in_list和seq_out_list形如
    ['五', '一', '劳', '动', '节', '是', '农', '历', '几', '时']
    ['B-datetime_date', 'I-datetime_date', 'I-datetime_date', 'I-datetime_date', 'I-datetime_date', 'O', 'O', 'O', 'O', 'O']
    '''
    assert type(seq_in_list)==type(seq_out_list)==list
    examples=[]
    if seq_out_list is not None:
        if type(seq_in_list[0])==list:
            assert type(seq_out_list[0])==list and len(seq_in_list[0])==len(seq_out_list[0])
        else:
            assert len(seq_in_list[0].split(' '))==len(seq_out_list[0].split(' '))

        for seq_in,seq_out in zip(seq_in_list,seq_out_list):
            examples.append(InputExample(seq_in=seq_in,seq_out=seq_out))
    else:
        for seq_in in seq_in_list:
            if type(seq_in)==list:
                seq_out=['O']*len(seq_in)
            else:
                assert isinstance(seq_in)==str
                seq_out=' '.join(['O']*len(seq_in.split(' ')))
            examples.append(InputExample(seq_in=seq_in,seq_out=seq_out))
    return examples

def convert_examples_to_features(examples, tokenizer, max_seq_len, label2id):
    '''
    label_ids在CLS、SEP、PAD位置上都是0，代表这些位置都是一个tag
    attention_mask仍然是考虑CLS和SEP的

    example的seq_in和seq_out都是列表
    '''
    features = []
    labels=[]
    special_token_nums=2#[CLS],[SEP]
    assert isinstance(examples[0],InputExample)#强制要求传进来的examples的element必须是InputExample类型

    max_len_this_batch=0
    for example in examples:
        length=len(example.seq_in)
        if length>max_len_this_batch:
            max_len_this_batch=length

    max_seq_len=min(max_seq_len,max_len_this_batch+special_token_nums)

    for example_index, example in enumerate(examples):
        tokens = []
        label_ids = []
        token_type_ids = []

        tokens.append("[CLS]")
        token_type_ids.append(0)
        label_ids.append(label2id["[PAD]"])#需要特别注意，只需要让CLS,SEP,PAD当作一个类别预测即可
        #没必要让[PAD]和[SEP]也作为类别，而且label2id['[PAD]']必须是0
        
        sentence_tokens = example.seq_in
        label_list = example.seq_out
        
        assert len(sentence_tokens)==len(label_list)
        if len(sentence_tokens)>=max_seq_len-special_token_nums:
            sentence_tokens=sentence_tokens[:max_seq_len-special_token_nums]
            label_list=label_list[:max_seq_len-special_token_nums]

        tokens += sentence_tokens
        token_type_ids += [0] * len(sentence_tokens)
        label_ids += [label2id[label] for label in label_list]

        tokens.append('[SEP]')
        token_type_ids.append(0)
        label_ids.append(label2id["[PAD]"])#只需要用一个tag来代表这三个特殊字符即可
        ##没必要让[PAD]和[SEP]也作为类别，而且label2id['[PAD]']必须是0

        seq_len = len(tokens)
        pad_len = max_seq_len - seq_len
        # print(tokens)
        input_ids = tokenizer.convert_tokens_to_ids(tokens)
        input_ids += [0] * pad_len
        attention_mask = [1] * seq_len + [0] * pad_len
        token_type_ids += [0] * pad_len
        label_ids += [label2id["[PAD]"]] * pad_len

        assert len(input_ids) == len(token_type_ids) == len(attention_mask) == max_seq_len == len(label_ids)

        features.append(InputFeatures(input_ids=input_ids, 
                                    token_type_ids=token_type_ids, 
                                    attention_mask=attention_mask,
                                    label_ids=label_ids))
        # if example_index < 5:
        #     logging.info("*** Example ***")
        #     logging.info("example_index: %s" % (example_index))
        #     logging.info("tokens: %s" % " ".join(tokens))
        #     logging.info("input_ids: %s" % " ".join([str(x) for x in input_ids]))
        #     logging.info("attention_mask: %s" % " ".join([str(x) for x in attention_mask]))
        #     logging.info("token_type_ids: %s" % " ".join([str(x) for x in token_type_ids]))
        #     logging.info("label_ids : %s" % " ".join([str(example.label)]))
    
    features=convert_to_tensor(features=features)
    return features



import logging
logging.basicConfig(level = logging.INFO,
                    format = '%(asctime)s - %(levelname)s - %(funcName)s - %(lineno)d : %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')


import torch
from torch import Tensor
from torch import nn
from torch import functional as F
from typing import Union, Tuple, List, Iterable, Dict
import os,sys
import json

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from log import logging
logger=logging.getLogger(__name__)

class LSTMLayer(nn.Module):
    def __init__(self, input_size, 
                        hidden_size, 
                        num_layers, 
                        bias=True, 
                        batch_first=True, 
                        dropout=0.0,
                        bidirectional=True):
        super(LSTMLayer,self).__init__()
        self.input_size=input_size
        self.hidden_size=hidden_size
        self.num_layers=num_layers
        self.bias = bias
        self.batch_first=batch_first
        self.dropout=dropout
        self.bidirectional=bidirectional
        self.lstmLayer=nn.LSTM(input_size=self.input_size,
                                hidden_size=self.hidden_size,
                                num_layers=self.num_layers,
                                bias=self.bias,
                                batch_first=self.batch_first,
                                dropout=self.dropout,
                                bidirectional=self.bidirectional)

    def forward(self, sequence_output, attention_mask):
        input_lengths=attention_mask.sum(1)
        packed_inputs_of_lstm=torch.nn.utils.rnn.pack_padded_sequence(input=sequence_output,lengths=input_lengths,batch_first=self.batch_first,enforce_sorted=False)
        sequence_output,_=self.lstmLayer(packed_inputs_of_lstm)
        sequence_output,_=torch.nn.utils.rnn.pad_packed_sequence(sequence_output,batch_first=self.batch_first)
        #sequence_output.size()==(bsa,max_seq_len,hidden_size*2)
        return sequence_output

    def get_config_dict(self):
        '''
        一定要有dict，这样才能初始化Model
        '''
        return {'input_size': self.input_size, 
                'hidden_size': self.hidden_size, 
                'bias': self.bias,
                'num_layers':self.num_layers,
                'batch_first':self.batch_first,
                'dropout':self.dropout,
                'bidirectional':self.bidirectional}

    def save(self,output_path):
        '''
        同时保存dict
        '''
        os.makedirs(output_path,exist_ok=True)
        with open(os.path.join(output_path, 'config.json'), 'w') as fOut:
            json.dump(self.get_config_dict(), fOut)

        torch.save(self.state_dict(),os.path.join(output_path, 'pytorch_model.bin'))
    
    @staticmethod
    def load(input_path):
        with open(os.path.join(input_path, 'config.json')) as fIn:
            config = json.load(fIn)

        model = LSTMLayer(**config)
        model.load_state_dict(torch.load(os.path.join(input_path, 'pytorch_model.bin'), map_location=torch.device('cpu')))
        return model


import torch

def convert_to_tensor(features):
    features_tensor={key:[] for key in features[0].__dict__.keys()}
    for feature in features:
        for key,value in feature.__dict__.items():
            features_tensor[key].append(value)
    features_tensor={key:torch.LongTensor(value) for key,value in features_tensor.items()}
    return features_tensor


import numpy as np
import os,json,sys
from typing import Dict, Sequence, Type, Callable, List, Optional
import torch
from torch import nn
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from modules.transformers import BertModel,BertConfig,BertTokenizer
from modules.MLP import MLP
from log import logging
logger=logging.getLogger(__name__)

'''
BertModel inherite BertForPretrainedModel inherite PreTrainedModel

In PreTrainedModel
@classmethod
def from_pretrained(cls,pretrained_model_name_or_path,*model_args,**kwargs):
    config=kwargs.pop('config',None)
    state_dict=kwargs.pop('state_dict',None)

    #The config will be loaded from pretrained_model_name_or_path/config.json if config not provided
    #instantiate model
    model=cls(config)
    if state_dict is None:
        state_dict=torch.load(pretrained_model_name_or_path/pytorch_model.bin,map_location='cpu')#也就是说模型最开始是加载到cpu上的
    missing_keys,unexpected_keys,error_msgs=[],[],[]

    def load(module: nn.Module, prefix=''):
        module._load_from_state_dict(state_dict,prefix,missing_keys,unexpected_keys,error_msgs)
        for name,child in module._modules.items():
            if child:
                load(child,prefix+name+'.')
    
    load(model,prefix='' if hasattr(model,'bert') else 'bert.')
    #prefix用来解决加载不同模型保存的pytorch_model.bin出现的keys不匹配问题
    if len(missing_keys) > 0:
        logger.info("Weights of {} not initialized from pretrained model: {}".format(
            model.__class__.__name__, missing_keys))
    if len(unexpected_keys) > 0:
        logger.info("Weights from pretrained model not used in {}: {}".format(
            model.__class__.__name__, unexpected_keys))
    return model
'''

class ClsHead(nn.Module):
    '''
    ClsHead不区分是单句子还是双句子，因为处理逻辑是一样的
    '''
    def __init__(self, model_path,
                num_labels,
                state_dict=None,
                is_finetune=False):
        super().__init__()
        self.num_labels=num_labels

        if is_finetune==False:
            logger.info("Loading model from {}, which is from huggingface model".format(model_path))
            self.load_huggingface_model(bert_model_path=model_path)
        else:
            self.load_finetuned_model(model_path=model_path)
            logger.info("Loading model from {}, which has been finetuned.".format(model_path))

        # bert_model_path=os.path.join(model_path,"BERT")#save的时候将BERT保存在model_path下的BERT文件夹中
        # self.config=BertConfig.from_pretrained(bert_model_path)
        # self.bert=BertModel.from_pretrained(bert_model_path)
        # self.tokenizer=BertTokenizer.from_pretrained(bert_model_path)

    def load_huggingface_model(self,bert_model_path):
        self.config=BertConfig.from_pretrained(bert_model_path)
        self.bert=BertModel.from_pretrained(bert_model_path)
        self.tokenizer=BertTokenizer.from_pretrained(bert_model_path)
        self.head_layer=MLP(in_features=self.config.hidden_size,out_features=self.num_labels)

    def load_finetuned_model(self,model_path):
        bert_save_path=os.path.join(model_path,"BERT")#save的时候将BERT保存在model_path下的BERT文件夹中
        self.config=BertConfig.from_pretrained(bert_save_path)
        self.bert=BertModel.from_pretrained(bert_save_path)
        self.tokenizer=BertTokenizer.from_pretrained(bert_save_path)

        head_save_path=os.path.join(model_path,'MLP')
        self.head_layer=MLP.load(input_path=head_save_path)

    def save(self,output_path):
        bert_save_path=os.path.join(output_path,"BERT")
        self.bert.save_pretrained(bert_save_path,save_config=False)#下面已经save，不用save两次，虽然没什么影响
        self.config.save_pretrained(bert_save_path)
        self.tokenizer.save_pretrained(bert_save_path)
        head_save_path=os.path.join(output_path,'MLP')
        self.head_layer.save(head_save_path)
    
    def forward(self,input_ids,attention_mask=None,token_type_ids=None,label_ids=None,output_all_encoded_layers=False):
        '''
        input_ids.size()==attention_mask.size()==token_type_ids.size()==position_ids.size()==(batch_size,seq_length)
        label_ids.size()==(batch_size,)
        '''
        (sequence_outputs,pooled_output)=self.bert(input_ids=input_ids,
                                                  token_type_ids=token_type_ids,
                                                   attention_mask=attention_mask,
                                                   output_all_encoded_layers=output_all_encoded_layers)
        #要注意到sequence_output[0]与pooled_output的区别在于pooled_output是经过一层tanh的
        assert len(pooled_output.size())==2 and pooled_output.size(1)==self.config.hidden_size
        logits=self.head_layer(pooled_output)#(batch_size,num_labels)
        return logits
        # predictions=torch.argmax(logits,dim=1)
        # if label_ids is not None:
        #     loss=nn.CrossEntropyLoss(reduction="mean")(input=logits,target=label_ids)
        #     accuracy=(predictions==label_ids).float().mean()
        #     return loss,accuracy
        # else:
        #     return logits,predictions


import importlib
from .transformers.optimization import get_cosine_with_hard_restarts_schedule_with_warmup,get_constant_schedule,get_linear_schedule_with_warmup,get_polynomial_decay_schedule_with_warmup,get_constant_schedule_with_warmup,get_cosine_schedule_with_warmup
from .transformers.optimization import AdamW,Adafactor

def get_scheduler(optimizer, scheduler: str, warmup_steps: int, t_total: int ):
    scheduler = scheduler.lower()
    if scheduler == 'constantlr':
        return get_constant_schedule(optimizer)
    elif scheduler == 'warmupconstant':
        return get_constant_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps)
    elif scheduler == 'warmuplinear':
        return get_linear_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total)
    elif scheduler == 'warmupcosine':
        return get_cosine_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total)
    elif scheduler == 'warmupcosinewithhardrestarts':
        return get_cosine_with_hard_restarts_schedule_with_warmup(optimizer, num_warmup_steps=warmup_steps, num_training_steps=t_total)
    else:
        raise ValueError("Unknown scheduler {}".format(scheduler))

def get_optimizer(model,optimizer_type,weight_decay,optimizer_params):
    param_optimizer = list(model.named_parameters())

    no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
    optimizer_grouped_parameters = [
        {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': weight_decay},
        {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
    ]

    if optimizer_type=='AdamW':
        optimizer=AdamW(optimizer_grouped_parameters, **optimizer_params)
    elif optimizer_type=='Adafactor':
        optimizer=Adafactor(optimizer_grouped_parameters, **optimizer_params)
    else:
        raise Exception("Unknown optimizer type")
    return optimizer



def fullname(o):
  """
  Gives a full name (package_name.class_name) for a class / object in Python. Will
  be used to load the correct classes from JSON files
  """

  module = o.__class__.__module__
  if module is None or module == str.__class__.__module__:
      return o.__class__.__name__  # Avoid reporting __builtin__
  else:
      return module + '.' + o.__class__.__name__

def import_from_string(dotted_path):
    '''
    如果输入是torch.nn.modules.activation.Tanh，那么module_path=torch.nn.modules.activation，class_name=Tanh
    对应的module是<module 'torch.nn.modules.activation' from '/root/miniconda3/envs/nlp_sr/lib/python3.6/site-packages/torch/nn/modules/activation.py'>
    返回的是torch.nn.modules.activation.Tanh

    如果输入是nlp_basictask.modules.transformers.BertModel
    对应的module是<module 'nlp_basictask.modules.transformers' from '/root/NLP_warehouse/nlp_basictask/modules/transformers/__init__.py'>
    返回的是nlp_basictask.modules.transformers.BertModel

    也就是说根据传进来的string返回对应的Model
    '''
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        msg = "%s doesn't look like a module path" % dotted_path
        raise ImportError(msg)
    
    try:
        module = importlib.import_module(dotted_path)
    except:
        module = importlib.import_module(module_path)
    
    try:
        return getattr(module,class_name)
    except AttributeError:
        msg = 'Module "%s" does not define a "%s" attribute/class' % (module_path, class_name)
        raise ImportError(msg)


# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HugginFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch BERT model."""

#from __future__ import absolute_import
#from __future__ import division
from __future__ import print_function

import os
import copy
import sys
import math
import tarfile
import tempfile
import shutil

import torch
from torch import nn
from .configuration_bert import BertConfig
#sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from .log import logging
logger = logging.getLogger(__name__)
#logger.info("加载BERT模型")

#CONFIG_NAME = 'bert_config.json'
CONFIG_NAME = 'config.json'#版本问题，这里进行了修改
WEIGHTS_NAME = 'pytorch_model.bin'

def gelu(x):
    """Implementation of the gelu activation function.
        For information: OpenAI GPT's gelu is slightly different (and gives slightly different results):
        0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
    """
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))


def swish(x):
    return x * torch.sigmoid(x)


ACT2FN = {"gelu": gelu, "relu": torch.nn.functional.relu, "swish": swish}

try:
    from apex.normalization.fused_layer_norm import FusedLayerNorm as BertLayerNorm
except ImportError:
    print("Better speed can be achieved with apex installed from https://www.github.com/nvidia/apex.")
    class BertLayerNorm(nn.Module):
        def __init__(self, hidden_size, eps=1e-12):
            """Construct a layernorm module in the TF style (epsilon inside the square root).
            """
            super(BertLayerNorm, self).__init__()
            self.weight = nn.Parameter(torch.ones(hidden_size))
            self.bias = nn.Parameter(torch.zeros(hidden_size))
            self.variance_epsilon = eps

        def forward(self, x):
            u = x.mean(-1, keepdim=True)
            s = (x - u).pow(2).mean(-1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.variance_epsilon)
            return self.weight * x + self.bias

def get_parameter_dtype(parameter):
    try:
        return next(parameter.parameters()).dtype
    except StopIteration:
        # For nn.DataParallel compatibility in PyTorch 1.5

        def find_tensor_attributes(module: nn.Module):
            tuples = [(k, v) for k, v in module.__dict__.items() if torch.is_tensor(v)]
            return tuples

        gen = parameter._named_members(get_members_fn=find_tensor_attributes)
        first_tuple = next(gen)
        return first_tuple[1].dtype

class BertEmbeddings(nn.Module):
    """Construct the embeddings from word, position and token_type embeddings.
    """
    def __init__(self, config):
        super(BertEmbeddings, self).__init__()
        self.word_embeddings = nn.Embedding(config.vocab_size, config.hidden_size)
        self.position_embeddings = nn.Embedding(config.max_position_embeddings, config.hidden_size)
        self.token_type_embeddings = nn.Embedding(config.type_vocab_size, config.hidden_size)

        # self.LayerNorm is not snake-cased to stick with TensorFlow model variable name and be able to load
        # any TensorFlow checkpoint file
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, input_ids, token_type_ids=None):
        seq_length = input_ids.size(1)
        position_ids = torch.arange(seq_length, dtype=torch.long, device=input_ids.device)
        position_ids = position_ids.unsqueeze(0).expand_as(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        words_embeddings = self.word_embeddings(input_ids)
        position_embeddings = self.position_embeddings(position_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embeddings = words_embeddings + position_embeddings + token_type_embeddings
        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class BertSelfAttention(nn.Module):
    def __init__(self, config):
        super(BertSelfAttention, self).__init__()
        if config.hidden_size % config.num_attention_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (config.hidden_size, config.num_attention_heads))
        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, hidden_states, attention_mask):
        mixed_query_layer = self.query(hidden_states)
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)

        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(mixed_key_layer)
        value_layer = self.transpose_for_scores(mixed_value_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
        attention_scores = attention_scores + attention_mask

        # Normalize the attention scores to probabilities.
        attention_probs = nn.Softmax(dim=-1)(attention_scores)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        context_layer = torch.matmul(attention_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)
        return context_layer


class BertSelfOutput(nn.Module):
    def __init__(self, config):
        super(BertSelfOutput, self).__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class BertAttention(nn.Module):
    def __init__(self, config):
        super(BertAttention, self).__init__()
        self.self = BertSelfAttention(config)
        self.output = BertSelfOutput(config)

    def forward(self, input_tensor, attention_mask):
        self_output = self.self(input_tensor, attention_mask)
        attention_output = self.output(self_output, input_tensor)
        return attention_output


class BertIntermediate(nn.Module):
    def __init__(self, config):
        super(BertIntermediate, self).__init__()
        self.dense = nn.Linear(config.hidden_size, config.intermediate_size)
        self.intermediate_act_fn = ACT2FN[config.hidden_act] \
            if isinstance(config.hidden_act, str) else config.hidden_act

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states


class BertOutput(nn.Module):
    def __init__(self, config):
        super(BertOutput, self).__init__()
        self.dense = nn.Linear(config.intermediate_size, config.hidden_size)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class BertLayer(nn.Module):
    def __init__(self, config):
        super(BertLayer, self).__init__()
        self.attention = BertAttention(config)
        self.intermediate = BertIntermediate(config)
        self.output = BertOutput(config)

    def forward(self, hidden_states, attention_mask):
        attention_output = self.attention(hidden_states, attention_mask)
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)
        return layer_output


class BertEncoder(nn.Module):
    def __init__(self, config):
        super(BertEncoder, self).__init__()
        layer = BertLayer(config)
        self.layer = nn.ModuleList([copy.deepcopy(layer) for _ in range(config.num_hidden_layers)])

    def forward(self, hidden_states, attention_mask, output_all_encoded_layers=True):
        all_encoder_layers = []
        for layer_module in self.layer:
            hidden_states = layer_module(hidden_states, attention_mask)
            if output_all_encoded_layers:
                all_encoder_layers.append(hidden_states)
        if not output_all_encoded_layers:
            all_encoder_layers.append(hidden_states)
        return all_encoder_layers


class BertPooler(nn.Module):
    def __init__(self, config):
        super(BertPooler, self).__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.activation = nn.Tanh()

    def forward(self, hidden_states):
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


class BertPredictionHeadTransform(nn.Module):
    def __init__(self, config):
        super(BertPredictionHeadTransform, self).__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.transform_act_fn = ACT2FN[config.hidden_act] \
            if isinstance(config.hidden_act, str) else config.hidden_act
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.transform_act_fn(hidden_states)
        hidden_states = self.LayerNorm(hidden_states)
        return hidden_states


class BertLMPredictionHead(nn.Module):
    def __init__(self, config, bert_model_embedding_weights):
        super(BertLMPredictionHead, self).__init__()
        self.transform = BertPredictionHeadTransform(config)

        # The output weights are the same as the input embeddings, but there is
        # an output-only bias for each token.
        self.decoder = nn.Linear(bert_model_embedding_weights.size(1),
                                 bert_model_embedding_weights.size(0),
                                 bias=False)
        self.decoder.weight = bert_model_embedding_weights
        self.bias = nn.Parameter(torch.zeros(bert_model_embedding_weights.size(0)))

    def forward(self, hidden_states):
        hidden_states = self.transform(hidden_states)
        hidden_states = self.decoder(hidden_states) + self.bias
        return hidden_states


class BertOnlyMLMHead(nn.Module):
    def __init__(self, config, bert_model_embedding_weights):
        super(BertOnlyMLMHead, self).__init__()
        self.predictions = BertLMPredictionHead(config, bert_model_embedding_weights)

    def forward(self, sequence_output):
        prediction_scores = self.predictions(sequence_output)
        return prediction_scores


class BertOnlyNSPHead(nn.Module):
    def __init__(self, config):
        super(BertOnlyNSPHead, self).__init__()
        self.seq_relationship = nn.Linear(config.hidden_size, 2)

    def forward(self, pooled_output):
        seq_relationship_score = self.seq_relationship(pooled_output)
        return seq_relationship_score


class BertPreTrainingHeads(nn.Module):
    def __init__(self, config, bert_model_embedding_weights):
        super(BertPreTrainingHeads, self).__init__()
        self.predictions = BertLMPredictionHead(config, bert_model_embedding_weights)
        self.seq_relationship = nn.Linear(config.hidden_size, 2)

    def forward(self, sequence_output, pooled_output):
        prediction_scores = self.predictions(sequence_output)
        seq_relationship_score = self.seq_relationship(pooled_output)
        return prediction_scores, seq_relationship_score


class BertPreTrainedModel(nn.Module):
    """ An abstract class to handle weights initialization and
        a simple interface for dowloading and loading pretrained models.
    """
    def __init__(self, config, *inputs, **kwargs):
        super(BertPreTrainedModel, self).__init__()
        if not isinstance(config, BertConfig):
            raise ValueError(
                "Parameter config in `{}(config)` should be an instance of class `BertConfig`. "
                "To create a model from a Google pretrained model use "
                "`model = {}.from_pretrained(PRETRAINED_MODEL_NAME)`".format(
                    self.__class__.__name__, self.__class__.__name__
                ))
        self.config = config

    def init_bert_weights(self, module):
        """ Initialize the weights.
        """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
        elif isinstance(module, BertLayerNorm):
            module.bias.data.normal_(mean=0.0, std=self.config.initializer_range)
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    @classmethod
    def from_pretrained(cls, pretrained_model_name, state_dict=None, cache_dir=None, *inputs, **kwargs):
        """
        Instantiate a PreTrainedBertModel from a pre-trained model file or a pytorch state dict.
        Download and cache the pre-trained model file if needed.

        Params:
            pretrained_model_name: either:
                - a str with the name of a pre-trained model to load selected in the list of:
                    . `bert-base-uncased`
                    . `bert-large-uncased`
                    . `bert-base-cased`
                    . `bert-base-multilingual`
                    . `bert-base-chinese`
                - a path or url to a pretrained model archive containing:
                    . `bert_config.json` a configuration file for the model
                    . `pytorch_model.bin` a PyTorch dump of a BertForPreTraining instance
            cache_dir: an optional path to a folder in which the pre-trained models will be cached.
            state_dict: an optional state dictionnary (collections.OrderedDict object) to use instead of Google pre-trained models
            *inputs, **kwargs: additional input for the specific Bert class
                (ex: num_labels for BertForSequenceClassification)
        """
#         if pretrained_model_name in PRETRAINED_MODEL_ARCHIVE_MAP:
#             archive_file = PRETRAINED_MODEL_ARCHIVE_MAP[pretrained_model_name]
#         else:
        archive_file = pretrained_model_name#这里修改，默认已经下载好了预训练模型
        # redirect to the cache, if necessary
#         try:
#             resolved_archive_file = cached_path(archive_file, cache_dir=cache_dir)
#         except FileNotFoundError:
#             logger.error(
#                 "Model name '{}' was not found in model name list ({}). "
#                 "We assumed '{}' was a path or url but couldn't find any file "
#                 "associated to this path or url.".format(
#                     pretrained_model_name,
#                     ', '.join(PRETRAINED_MODEL_ARCHIVE_MAP.keys()),
#                     archive_file))
#             return None
        resolved_archive_file=archive_file#这里进行了修改，因为在cached_path中，如果传进去的archive_file存在的话，直接return
        if resolved_archive_file == archive_file:
            logger.info("loading bert model file {}".format(archive_file))
        else:
            logger.info("loading archive file {} from cache at {}".format(
                archive_file, resolved_archive_file))
        tempdir = None
        if os.path.isdir(resolved_archive_file):
            serialization_dir = resolved_archive_file
        else:
            # Extract archive to temp dir
            tempdir = tempfile.mkdtemp()
            logger.info("extracting archive file {} to temp dir {}".format(
                resolved_archive_file, tempdir))
            with tarfile.open(resolved_archive_file, 'r:gz') as archive:
                archive.extractall(tempdir)
            serialization_dir = tempdir
        # Load config
        config_file = os.path.join(serialization_dir, CONFIG_NAME)
        config = BertConfig.from_json_file(config_file)
        #logger.info("Model config {}".format(config))
        logger.info("BertConfig has been loaded from {}".format(config_file))
        # Instantiate model.
        model = cls(config, *inputs, **kwargs)
        if state_dict is None:
            weights_path = os.path.join(serialization_dir, WEIGHTS_NAME)
            state_dict = torch.load(weights_path)

        old_keys = []
        new_keys = []
        for key in state_dict.keys():
            new_key = None
            if 'gamma' in key:
                new_key = key.replace('gamma', 'weight')
            if 'beta' in key:
                new_key = key.replace('beta', 'bias')
            if new_key:
                old_keys.append(key)
                new_keys.append(new_key)
        for old_key, new_key in zip(old_keys, new_keys):
            state_dict[new_key] = state_dict.pop(old_key)

        missing_keys = []
        unexpected_keys = []
        error_msgs = []
        # copy state_dict so _load_from_state_dict can modify it
        metadata = getattr(state_dict, '_metadata', None)
        state_dict = state_dict.copy()
        if metadata is not None:
            state_dict._metadata = metadata

        def load(module, prefix=''):
            local_metadata = {} if metadata is None else metadata.get(prefix[:-1], {})
            module._load_from_state_dict(
                state_dict, prefix, local_metadata, True, missing_keys, unexpected_keys, error_msgs)
            for name, child in module._modules.items():
                if child is not None:
                    load(child, prefix + name + '.')
        load(model, prefix='' if hasattr(model, 'bert') else 'bert.')
        if len(missing_keys) > 0:
            logger.info("Weights of {} not initialized from pretrained model: {}".format(
                model.__class__.__name__, missing_keys))
        if len(unexpected_keys) > 0:
            logger.info("Weights from pretrained model not used in {}: {}".format(
                model.__class__.__name__, unexpected_keys))
        if tempdir:
            # Clean up temp dir
            shutil.rmtree(tempdir)
        return model
    
    def save_pretrained(self,save_directory,
                            save_config=True,
                            state_dict=None,
                            save_function=torch.save,
                            ):
        def unwrap_model(model: nn.Module) -> nn.Module:
            """
            Recursively unwraps a model from potential containers (as used in distributed training).

            Args:
                model (:obj:`torch.nn.Module`): The model to unwrap.
            """
            # since there could be multiple levels of wrapping, unwrap recursively
            if hasattr(model, "module"):
                return unwrap_model(model.module)
            else:
                return model
        if os.path.isfile(save_directory):
            logger.info(f"Provided path ({save_directory}) should be a directory, not a file")
            return 
        os.makedirs(save_directory,exist_ok=True)
        model_to_save = unwrap_model(self)
        dtype = get_parameter_dtype(model_to_save)
        model_to_save.config.torch_dtype = str(dtype).split(".")[1]
        model_to_save.config.architectures = [model_to_save.__class__.__name__]

        if save_config:
            model_to_save.config.save_pretrained(save_directory)
        
        if state_dict is None:
            state_dict = model_to_save.state_dict()

        output_model_file = os.path.join(save_directory, WEIGHTS_NAME)
        state_dict_keys_list=list(state_dict.keys())
        if state_dict_keys_list[0].split('.')!='bert':
            state_dict={'bert.'+key:value for key,value in state_dict.items()}
        save_function(state_dict, output_model_file)
        logger.info(f"Model weights saved in {output_model_file}")


class BertModel(BertPreTrainedModel):
    """BERT model ("Bidirectional Embedding Representations from a Transformer").

    Params:
        config: a BertConfig class instance with the configuration to build a new model

    Inputs:
        `input_ids`: a torch.LongTensor of shape [batch_size, sequence_length]
            with the word token indices in the vocabulary(see the tokens preprocessing logic in the scripts
            `extract_features.py`, `run_classifier.py` and `run_squad.py`)
        `token_type_ids`: an optional torch.LongTensor of shape [batch_size, sequence_length] with the token
            types indices selected in [0, 1]. Type 0 corresponds to a `sentence A` and type 1 corresponds to
            a `sentence B` token (see BERT paper for more details).
        `attention_mask`: an optional torch.LongTensor of shape [batch_size, sequence_length] with indices
            selected in [0, 1]. It's a mask to be used if the input sequence length is smaller than the max
            input sequence length in the current batch. It's the mask that we typically use for attention when
            a batch has varying length sentences.
        `output_all_encoded_layers`: boolean which controls the content of the `encoded_layers` output as described below. Default: `True`.

    Outputs: Tuple of (encoded_layers, pooled_output)
        `encoded_layers`: controled by `output_all_encoded_layers` argument:
            - `output_all_encoded_layers=True`: outputs a list of the full sequences of encoded-hidden-states at the end
                of each attention block (i.e. 12 full sequences for BERT-base, 24 for BERT-large), each
                encoded-hidden-state is a torch.FloatTensor of size [batch_size, sequence_length, hidden_size],
            - `output_all_encoded_layers=False`: outputs only the full sequence of hidden-states corresponding
                to the last attention block of shape [batch_size, sequence_length, hidden_size],
        `pooled_output`: a torch.FloatTensor of size [batch_size, hidden_size] which is the output of a
            classifier pretrained on top of the hidden state associated to the first character of the
            input (`CLF`) to train on the Next-Sentence task (see BERT's paper).

    Example usage:
    ```python
    # Already been converted into WordPiece token ids
    input_ids = torch.LongTensor([[31, 51, 99], [15, 5, 0]])
    input_mask = torch.LongTensor([[1, 1, 1], [1, 1, 0]])
    token_type_ids = torch.LongTensor([[0, 0, 1], [0, 1, 0]])

    config = modeling.BertConfig(vocab_size_or_config_json_file=32000, hidden_size=768,
        num_hidden_layers=12, num_attention_heads=12, intermediate_size=3072)

    model = modeling.BertModel(config=config)
    all_encoder_layers, pooled_output = model(input_ids, token_type_ids, input_mask)
    ```
    """
    def __init__(self, config):
        super(BertModel, self).__init__(config)
        self.embeddings = BertEmbeddings(config)
        self.encoder = BertEncoder(config)
        self.pooler = BertPooler(config)
        self.apply(self.init_bert_weights)

    def forward(self, input_ids, token_type_ids=None, attention_mask=None, output_all_encoded_layers=True):
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        # We create a 3D attention mask from a 2D tensor mask.
        # Sizes are [batch_size, 1, 1, to_seq_length]
        # So we can broadcast to [batch_size, num_heads, from_seq_length, to_seq_length]
        # this attention mask is more simple than the triangular masking of causal attention
        # used in OpenAI GPT, we just need to prepare the broadcast dimension here.
        extended_attention_mask = attention_mask.unsqueeze(1).unsqueeze(2)

        # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
        # masked positions, this operation will create a tensor which is 0.0 for
        # positions we want to attend and -10000.0 for masked positions.
        # Since we are adding it to the raw scores before the softmax, this is
        # effectively the same as removing these entirely.
        extended_attention_mask = extended_attention_mask.to(dtype=next(self.parameters()).dtype) # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0

        embedding_output = self.embeddings(input_ids, token_type_ids)
        encoded_layers = self.encoder(embedding_output,
                                      extended_attention_mask,
                                      output_all_encoded_layers=output_all_encoded_layers)
        sequence_output = encoded_layers[-1]
        pooled_output = self.pooler(sequence_output)
        if not output_all_encoded_layers:
            encoded_layers = encoded_layers[-1]
        return encoded_layers, pooled_output


import json
from logging import log
from nlp_basictasks.modules.transformers.modeling_bert import BertModel
from nlp_basictasks.modules.transformers.tokenization_bert import BertTokenizer
import numpy as np
import os,sys
from typing import Dict, Sequence, Text, Type, Callable, List, Optional, Union
from tqdm import tqdm

import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm, trange
from tensorboardX import SummaryWriter
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
#print(sys.path)

from log import logging

from modules.utils import get_optimizer,get_scheduler
from heads import SoftmaxLossHead
from modules.Pooling import Pooling
from modules.MLP import MLP
from readers.sts import convert_examples_to_features,convert_sentences_to_features
from .utils import batch_to_device,eval_during_training

logger=logging.getLogger(__name__)

class sts():
    def __init__(self,model_path,
                max_seq_length:int = 32,
                head_name:str = 'softmaxloss',
                head_config: dict = None,
                device:str = None,
                state_dict=None,
                is_finetune=False,
                tensorboard_logdir = None):
        '''
        head_name用来指明使用哪一个loss，包括softmaxloss和各种基于对比学习的loss
        head_config用来给出具体loss下所需要的参数，
        '''
        self.head_name=head_name
        self.head_config=head_config
        self.max_seq_lenth=max_seq_length
        if self.head_name=='softmaxloss':
            if self.head_config is None:
                self.head_config={'concatenation_sent_rep':True,
                                'concatenation_sent_difference':True,
                                'pooling_mode_mean_tokens':True}
            label2id={'0':0,'1':1}
            self.model=SoftmaxLossHead(model_path=model_path,
                                        num_labels=len(label2id),
                                        is_finetune=is_finetune,
                                        **self.head_config)
        
        else:
            raise Exception("Unknown loss {} for sts".format(head_name))

        if tensorboard_logdir!=None:
            self.tensorboard_writer=SummaryWriter(tensorboard_logdir)
        else:
            self.tensorboard_writer=None
        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"
            logging.info("Use pytorch device: {}".format(device))
        self._target_device = torch.device(device)
        self.model.to(self._target_device)

    def smart_batching_collate(self,batch):
        features_of_a,features_of_b,labels=convert_examples_to_features(examples=batch,tokenizer=self.model.tokenizer,max_seq_len=self.max_seq_lenth)
        return features_of_a,features_of_b,labels

    def encode(self, sentences,
               batch_size: int = 32,
               show_progress_bar: bool = None,
               output_value: str = 'sentence_embedding',
               convert_to_numpy: bool = True,
               convert_to_tensor: bool = False,
               device: str = None,
               normalize_embeddings: bool = False):
        '''
        传进来的sentences只能是single_batch
        '''
        self.model.eval()
        input_is_string=False
        if type(sentences)==str:
            input_is_string=True
            sentences=[sentences]
        if convert_to_tensor:
            convert_to_numpy = False
        if device is None:
            device = self._target_device

        all_embeddings = []
        length_sorted_idx = np.argsort([-len(sen) for sen in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]
        for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
            sentences_batch = sentences_sorted[start_index:start_index+batch_size]
            features=convert_sentences_to_features(sentences_batch,tokenizer=self.model.tokenizer)
            features=batch_to_device(features,device)
            with torch.no_grad():
                embeddings=self.model(sentence_features_of_1=features,sentence_features_of_2=None,encode_pattern=True)
                embeddings=embeddings.detach()
                if normalize_embeddings:
                    embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)
                    if convert_to_numpy:
                        embeddings = embeddings.cpu()
                all_embeddings.extend(embeddings)
        
        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]
        if convert_to_tensor:
            all_embeddings = torch.stack(all_embeddings)
        elif convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])

        if input_is_string:
            all_embeddings = all_embeddings[0]#一个hidden_size维度的vector

        return all_embeddings

    def fit(self,
            train_dataloader,
            evaluator = None,
            epochs: int = 1,
            loss_fct = nn.CrossEntropyLoss(),
            scheduler: str = 'WarmupLinear',
            warmup_proportion: float = 0.1,
            optimizer_type = 'AdamW',
            optimizer_params: Dict[str, object] = {'lr': 2e-5},
            weight_decay: float = 0.01,
            evaluation_steps = None,
            output_path: str = None,
            save_best_model: bool = True,
            max_grad_norm: float = 1,
            use_amp: bool = False,
            callback: Callable[[float, int, int], None] = None,
            show_progress_bar: bool = True,
            early_stop_patience = 10,
            print_loss_step: Optional[int] = None,
            output_all_encoded_layers: bool = False
            ):

        train_dataloader.collate_fn=self.smart_batching_collate
        if print_loss_step==None:
            print_loss_step=len(train_dataloader)//5
        if evaluator is not None and evaluation_steps==None:
            evaluation_steps=len(train_dataloader)//2
        #一个epoch下打印5次loss，评估2次
        logger.info("一个epoch 下，每隔{}个step会输出一次loss，每隔{}个step会评估一次模型".format(print_loss_step,evaluation_steps))
        if use_amp:
            from torch.cuda.amp import autocast
            scaler = torch.cuda.amp.GradScaler()

        self.model.to(self._target_device)#map_location='cpu'

        if output_path is not None:
            os.makedirs(output_path, exist_ok=True)

        self.best_score = -9999999
        num_train_steps = int(len(train_dataloader) * epochs)
        warmup_steps = num_train_steps*warmup_proportion

        optimizer = get_optimizer(model=self.model,optimizer_type=optimizer_type,weight_decay=weight_decay,optimizer_params=optimizer_params)
        scheduler = get_scheduler(optimizer, scheduler=scheduler, warmup_steps=warmup_steps, t_total=num_train_steps)

        global_step=0
        skip_scheduler = False
        patience = early_stop_patience        
        for epoch in trange(epochs, desc="Epoch", disable=not show_progress_bar):
            training_steps=0
            training_loss=0.0

            self.model.zero_grad()
            self.model.train()

            for train_step,(features_of_a,features_of_b,labels) in tqdm(enumerate(train_dataloader)):
                features_of_a=batch_to_device(features_of_a,target_device=self._target_device)
                features_of_b=batch_to_device(features_of_b,target_device=self._target_device)
                features={'sentence_features_of_1':features_of_a,
                        'sentence_features_of_2':features_of_b}
                labels=labels.to(self._target_device)
                features['output_all_encoded_layers']=output_all_encoded_layers
                #print(features.keys(),features["input_ids"].size())
                if use_amp:
                    with autocast():
                        logits=self.model(**features)
                        loss_value = loss_fct(logits, labels)

                    scale_before_step = scaler.get_scale()
                    scaler.scale(loss_value).backward()
                    scaler.unscale_(optimizer)
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                    scaler.step(optimizer)
                    scaler.update()

                    skip_scheduler = scaler.get_scale() != scale_before_step
                else:
                    logits=self.model(**features)
                    loss_value=loss_fct(logits,labels)#CrossEntropyLoss
                    training_loss+=loss_value.item()
                    if print_loss_step!=None and train_step>0 and train_step%print_loss_step == 0:
                        training_loss/=print_loss_step
                        logging.info("Epoch : {}, train_step : {}/{}, loss_value : {} ".format(epoch,train_step*(epoch+1),num_train_steps,training_loss))
                        training_loss=0.0
                    if self.tensorboard_writer is not None:
                        self.tensorboard_writer.add_scalar(f"train_loss",loss_value.item(),global_step=global_step)
                    loss_value.backward()
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                    optimizer.step()

                optimizer.zero_grad()
                if not skip_scheduler:
                    scheduler.step()

                training_steps += 1
                global_step += 1

                if evaluation_steps>0 and evaluator is not None and (train_step) % evaluation_steps == 0 :
                    eval_score=self._eval_during_training(evaluator, output_path, epoch, training_steps, callback)
                    if self.tensorboard_writer is not None:
                        self.tensorboard_writer.add_scalar(f"eval_score",float(eval_score),global_step=global_step)
                    if eval_score > self.best_score:
                        if save_best_model:
                            self.model.save(output_path=output_path)
                            logging.info(("In epoch {}, training_step {}, the eval score is {}, previous eval score is {}, model has been saved in {}".format(epoch,train_step*(epoch+1),eval_score,self.best_score,output_path)))
                            self.best_score=eval_score
                    else:
                        patience-=1
                        logging.info(f"No improvement over previous best eval score ({eval_score:.6f} vs {self.best_score:.6f}), patience = {patience}")
                        if patience==0:
                            logging.info("Run our of patience, early stop!")
                            return
                
                    self.model.zero_grad()
                    self.model.train()

    def _eval_during_training(self, evaluator, output_path,  epoch, steps, callback):
        if evaluator is not None:
            score_and_auc = evaluator(self, output_path=output_path, epoch=epoch, steps=steps)
            if callback is not None:
                callback(score_and_auc, epoch, steps)
            return score_and_auc
        return None



class SimilarityRetrieve(nn.Module):
    def __init__(self,bert_model_path,
                pooling_model_path=None,
                pooling_config={'pooling_mode_cls_token':True},
                mlp_model_path=None,
                max_seq_length:int = 32,
                device:str = None):
        '''
        
        '''
        super(SimilarityRetrieve,self).__init__()
        #logger.info("Bert model path and pooling model path is required")
        self.max_seq_lenth=max_seq_length
        if not os.path.exists(os.path.join(bert_model_path,'pytorch_model.bin')):
            raise Exception('Not found pytorch_model.bin in bert model path : {}'.format(bert_model_path))
        else:
            assert os.path.exists(os.path.join(bert_model_path,'config.json'))
            self.bert=BertModel.from_pretrained(bert_model_path)
            self.tokenizer=BertTokenizer.from_pretrained(bert_model_path)
            pooling_config.update({'word_embedding_dimension':self.bert.config.hidden_size})
        
        if pooling_model_path==None or not os.path.exists(os.path.join(pooling_model_path,'config.json')):
            self.pooling=Pooling(**pooling_config)
        else:
            self.pooling=Pooling.load(pooling_model_path)
        
        if mlp_model_path is not None:
            logger.info("The MLP model path is not empty, which means you will project the vector after pooling")
            if not os.path.exists(os.path.join(mlp_model_path,'pytorch_model.bin')):
                raise Exception('Not found pytorch_model.bin in MLP model path : {}'.format(mlp_model_path))
            else:
                assert os.path.exists(os.path.join(mlp_model_path,'config.json'))
                self.mlp=MLP.load(mlp_model_path)

        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"
            logger.info("Use pytorch device: {}".format(device))
        self._target_device=device
        self.to(device)

    def smart_batching_collate(self,batch):
        features_of_a,features_of_b,labels=convert_examples_to_features(examples=batch,tokenizer=self.model.tokenizer,max_seq_len=self.max_seq_lenth)
        return features_of_a,features_of_b,labels

    def encode(self, sentences,
               batch_size: int = 32,
               show_progress_bar: bool = None,
               output_value: str = 'sentence_embedding',
               convert_to_numpy: bool = True,
               convert_to_tensor: bool = False,
               device: str = None,
               normalize_embeddings: bool = False,
               output_all_encoded_layers=False):
        '''
        传进来的sentences只能是single_batch
        '''
        self.eval()
        input_is_string=False
        if type(sentences)==str:
            input_is_string=True
            sentences=[sentences]
        if convert_to_tensor:
            convert_to_numpy = False
        if device is None:
            device = self._target_device

        all_embeddings = []
        length_sorted_idx = np.argsort([-len(sen) for sen in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]
        for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
            sentences_batch = sentences_sorted[start_index:start_index+batch_size]
            features=convert_sentences_to_features(sentences_batch,tokenizer=self.tokenizer)
            features=batch_to_device(features,device)
            with torch.no_grad():
                embeddings=self.encoding_feature_to_embedding(sentence_features=features,output_all_encoded_layers=output_all_encoded_layers)
                embeddings=embeddings.detach()
                if normalize_embeddings:
                    embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)
                    if convert_to_numpy:
                        embeddings = embeddings.cpu()
                all_embeddings.extend(embeddings)
        
        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]
        if convert_to_tensor:
            all_embeddings = torch.stack(all_embeddings)
        elif convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])

        if input_is_string:
            all_embeddings = all_embeddings[0]#一个hidden_size维度的vector

        return all_embeddings

    def encoding_feature_to_embedding(self,sentence_features,output_all_encoded_layers=False):
        '''
        each_features like : {'input_ids':tensor,'attention_mask':tensor,'token_type_ids':tensor},
        input_ids.size()==attention_mask.size()==token_type_ids.size()==position_ids.size()==(batch_size,seq_length)
        label_ids.size()==(batch_size,)
        '''
        #只有在encode模式下的single_batch才是有意义的，不然如果不是encode模式，只传入一个句子，有没有标签，无法返回任何值
        batch_size,seq_len_1=sentence_features['input_ids'].size()

        input_ids=sentence_features['input_ids']
        token_type_ids=sentence_features['token_type_ids']
        attention_mask=sentence_features['attention_mask']

        (sequence_outputs,pooler_output)=self.bert(input_ids=input_ids,
                                                token_type_ids=token_type_ids,
                                                attention_mask=attention_mask,
                                                output_all_encoded_layers=output_all_encoded_layers)
        #要注意到sequence_output[0]与pooled_output的区别在于pooler_output是经过一层tanh的
        if output_all_encoded_layers:
            all_layer_embeddings=sequence_outputs
            token_embeddings=sequence_outputs[-1]
        else:
            all_layer_embeddings=None
            token_embeddings=sequence_outputs

        cls_token_embeddings=pooler_output
        sentence_embedding=self.pooling(token_embeddings=token_embeddings,
                                            cls_token_embeddings=cls_token_embeddings,
                                            attention_mask=attention_mask,
                                            all_layer_embeddings=all_layer_embeddings)
        assert sentence_embedding.size()==(batch_size,self.pooling.pooling_output_dimension)
        return sentence_embedding


import torch

def batch_to_device(batch,target_device):
    """
    send a pytorch batch to a device (CPU/GPU)
    """
    for key in batch:
        if isinstance(batch[key], torch.Tensor):
            batch[key] = batch[key].to(target_device)
    return batch

def eval_during_training(model, evaluator, output_path,  epoch, steps, callback):
    if evaluator is not None:
        score_and_auc = evaluator(model, output_path=output_path, epoch=epoch, steps=steps)
        if callback is not None:
            callback(score_and_auc, epoch, steps)
        return score_and_auc
    return None



import torch
from torch import Tensor
from torch import nn
from torch import functional as F
from typing import Union, Tuple, List, Iterable, Dict
import os,sys
import json

from .utils import fullname,import_from_string
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from log import logging
logger=logging.getLogger(__name__)

class MLP(nn.Module):
    def __init__(self, in_features, 
                        out_features, 
                        bias = True, 
                        activation_function=nn.Tanh(), 
                        init_weight: Tensor = None, 
                        init_bias: Tensor = None):
        super(MLP,self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.bias = bias
        self.activation_function = activation_function
        self.linear = nn.Linear(in_features, out_features, bias=bias)

        if init_weight is not None:
            self.linear.weight = nn.Parameter(init_weight)

        if init_bias is not None:
            self.linear.bias = nn.Parameter(init_bias)

    def forward(self, features):
        '''
        The output shape is like features.shape except last dim
        '''
        return self.linear(features)

    def get_config_dict(self):
        '''
        一定要有dict，这样才能初始化Model
        '''
        return {'in_features': self.in_features, 'out_features': self.out_features, 'bias': self.bias, 'activation_function': fullname(self.activation_function)}

    def save(self,output_path):
        '''
        同时保存dict
        '''
        os.makedirs(output_path,exist_ok=True)
        with open(os.path.join(output_path, 'config.json'), 'w') as fOut:
            json.dump(self.get_config_dict(), fOut)

        torch.save(self.state_dict(),os.path.join(output_path, 'pytorch_model.bin'))
    
    @staticmethod
    def load(input_path):
        with open(os.path.join(input_path, 'config.json')) as fIn:
            config = json.load(fIn)

        config['activation_function'] = import_from_string(config['activation_function'])()
        #上一行是因为激活函数是函数，不是整数，所以要给出位置
        model = MLP(**config)
        model.load_state_dict(torch.load(os.path.join(input_path, 'pytorch_model.bin'), map_location=torch.device('cpu')))
        return model